#!/usr/bin/python

# $Id: bounce.py,v 1.16 2004/06/03 06:28:35 brinkman Exp $
# (c) 2004, Peter Brinkmann (brinkman@math.uiuc.edu)
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation (http://www.gnu.org/copyleft/gpl.html).

import sys
from math import *
from PySZG import *


class Scene:

	def __init__(self,scene):
		self.scene=scene

	def update(self,g=arVector3(0,-9.81,0),dt=0.005):
		for x in self.scene:
			x.interact(self.scene,g,dt)


class BounceObj:
	def interact(self,lst,g,dt):
		pass

class Ball(BounceObj):

	def __init__(self,x=arVector3(0,0,0),v=arVector3(0,0,0),
				r=1.0,m=1.0,k=500.0,ancestor='world'):
		self.x=x	# location
		self.v=v	# velocity
		self.m=m	# mass
		self.r=r	# radius
		self.k=k	# spring constant
		self.ID=dgTransform(`self`,ancestor,arMatrix4())
		self.mesh=arSphereMesh(40)
		self.mesh.attachMesh(`self`+'mesh',`self`)
		self.shape=ar_scaleMatrix(self.r)
		self.force=arVector3()

	def bounce(self,x):
		# implements a simple model of bouncing balls: if flattened by
		# dx, they'll push back with a force equal to k*dx (Hooke's law)
		if isinstance(x,Ball):
			self.bounceOffBall(x)
		elif isinstance(x,Wall):
			self.bounceOffWall(x)
		else:
			raise "Don't know how to interact with "+`x`

	def bounceOffWall(self,w):
		d=(self.x-w.p)%w.n		# signed distance from wall w
		if d>=self.r:
			return
		if d<=0:
			# keep balls from tunneling through walls
			self.v-=2*(self.v%w.n)*w.n
		else:
			self.force+=(self.r-abs(d))*(d/abs(d))*self.k*w.n
		q=abs(d)/self.r
		self.shape=self.shape*w.chBasisInv*ar_scaleMatrix(
						q,sqrt(1/q),sqrt(1/q))*w.chBasis
		# flatten the sphere appropriately, while preserving volume

	def bounceOffBall(self,b):
		v=self.x-b.x
		if not(self is b) and abs(v)<self.r+b.r:
			s=(self.r+b.r-abs(v))*b.k/(self.k+b.k)
			v/=abs(v)
			self.force+=v*s*self.k
			q=(self.r-s)/self.r
			chBasis=ar_rotateVectorToVector(v,arVector3(1,0,0))
			self.shape=self.shape*chBasis.invert()*ar_scaleMatrix(
							q,sqrt(1/q),sqrt(1/q))*chBasis

	def interact(self,lst,g,dt):
		self.shape=ar_scaleMatrix(self.r)
		self.force=arVector3()
		for x in lst:
			self.bounce(x)
		self.v+=(self.force/self.m+g)*dt
		self.x+=self.v*dt
		dgTransform(self.ID,ar_translationMatrix(self.x)*self.shape)


class Wall(BounceObj):

	def __init__(self,p=arVector3(0,0,0),n=arVector3(1,0,0)):
		self.p=p			# point on wall
		self.n=n/abs(n)		# normal vector of wall
		self.chBasis=ar_rotateVectorToVector(self.n,arVector3(1,0,0))
		self.chBasisInv=self.chBasis.invert()


fw=arDistSceneGraphFramework()
if not fw.init(sys.argv):
	raise 'Unable to initialize framework.'
ifo=arInterfaceObject()
ifo.setInputDevice(fw.getInputDevice())

# configure stereo view
fw.setEyeSpacing(6.0/(2.54*12.0))
fw.setClipPlanes(0.3,1000.0)
fw.setUnitConversion(1.0)

# some lights...
dgLight('light0',fw.getNavNodeName(),0,arVector4(0,0,1,0),arVector3(1,1,1))
dgLight('light1',fw.getNavNodeName(),1,arVector4(0,0,-1,0),arVector3(1,1,1))
dgLight('light2',fw.getNavNodeName(),2,arVector4(0,-1,0,0),arVector3(1,1,1))
dgLight('light3',fw.getNavNodeName(),3,arVector4(0,1,0,0),arVector3(1,1,1))

wID=dgTransform('world',fw.getNavNodeName(),arMatrix4())

sc=Scene([
	Ball(arVector3(0,0,0),arVector3(5,4,3),1.2,.5,1200.0),
	Ball(arVector3(1,1,-1),arVector3(1,1,2),1.0,2.0,3000.0),
	Wall(arVector3(0,-3,0),arVector3(0,1,0)),
	Wall(arVector3(0,3,0),arVector3(0,-1,0)),
	Wall(arVector3(-3,0,0),arVector3(1,0,0)),
	Wall(arVector3(3,0,0),arVector3(-1,0,0)),
	Wall(arVector3(0,0,-3),arVector3(0,0,1)),
	Wall(arVector3(0,0,2),arVector3(0,0,-1))
])

if not fw.start():
	raise 'Unable to start framework.'
if not ifo.start():
	raise 'Unable to start interface.'

if __name__=='__main__':
	import time
	print 'Use dkill or press Ctrl-C to quit.'
	g=arVector3(0,-9.81/.305,0)
	ifo.setNavMatrix(ar_translationMatrix(0,5,-5))
	while 1:
		ar_setNavMatrix(ifo.getNavMatrix().invert())
		fw.loadNavMatrix()
		objmat=ifo.getObjectMatrix()
		dgTransform(wID,objmat)
		fw.setViewer()
		g=objmat.invert()*arVector3(0,-9.81/.305,0)
		sc.update(g)
		time.sleep(0.01)
